#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Helpers and functions to create subcommands.


= Subcommands =

A subcommand can be any object implementing the following protocol:

    Required names:

        - help: the command line help of the subcommand.

        - get_description(namespace): a callable that receives a namespace
          and must return the description of the subcommand as a string that
          will be printed during interactive or dry run execution.
          Note that this name must only be defined if the subcommand supports
          interactive or dry run (see below in "Optional names").
          Also note that the namespace is passed so that the callable can
          retrieve data from it, but the callable itself should not mutate the
          namespace.

        - run(namespace): a callable actually executing the subcommand.
          Again, the namespace should not be mutataed by this callable.

    Optional names:

        - add_arguments(parser): a callable that receives a parser and can
          add subcommand specific arguments to that parser.

        - epilog: a string representing text to be added at the end of the
          auto-generated help message. The text can contain `%(prog)s` as a
          placeholder to current subcommand name, e.g 'lpsetup init-target'.

        - has_dry_run: True if the subcommand supports dry run execution,
          False otherwise. If not defined, False is considered.

        - has_interactive_run: True if the subcommand supports interactive
          execution, False otherwise. If not defined, the subcommand is
          considered non-interactive.

        - prepare_namespace(namespace): a callable that receives a namespace
          object. A subcommand can define this callable to further initialize
          or validate the namespace as needed. A ValidationError can be
          raised by *prepare_namespace* to stop the subcommand execution.
          This is the only place where the namespace can be mutated.

        - root_required: True if the subcommand requires the user to be root,
          False otherwise. If not defined, False is considered.

However, this module provides two classes that can be used to automate the
process in several ways.


== BaseSubCommand ==

This base class implements:

    - help: as None.

    - prepare_namespace(namespace):

        This base subcommand introduces the concept of handlers. Handlers are
        callables receiving the current namespace. They can manipulate the
        namespace adding or changing names. They can also validate the
        namespace, e.g. when the validation depends on multiple args.
        If something is wrong with the arguments, a handler should raise a
        *ValidationError*.

        Subcommands can store handlers in the *handlers* attribute as a
        sequence, or can dynamically set up handlers overriding the
        *get_handlers* method. By default, *get_handlers* returns
        *self.handlers*.

        In any case, the *prepare_namespace* implementation of this class
        calls the handlers in the order they are provided by *get_handlers*.

    - get_description(namespace): returning an empty string.

All the required names in the subcommands protocol are implemented by
*BaseSubCommand* but *run*, that must be overridden by subclasses to
do actual stuff.


== StepsBasedSubCommand ==

Instances of this class (subclass of *BaseSubCommand*) can take advantage of
the integrated "steps" machinery. The *run* method is implemented to execute
steps in order. The main advantage of basing a subcommand on steps is the
possibility of executing only a subset of the subcommand, including or
skipping some steps. This allows more granularity when the subcommand is run.
Note that this base class is a subclass of *BaseSubCommand*.
In particular, this base class implements:

    - steps:

        a sequence of *(callable, arg-names, filter-callable)* where:

        - callable is a callable step that will be run passing collected *args*
        - *arg-names* are the names of namespace attributes, used to retrieve
          the corresponding values and generate a tuple of *args* to be
          passed to the step
        - *filter-callable* is an OPTIONAL function that takes the namespace
          and returns True if the step must be run, False otherwise.
          If not provided, the step is run.

        Example: run *mystep*, passing *foo* and *bar* values,
        only if the namespace contains 'foo' and 'foo' is True::

            def my_step(foo, bar):
                pass

            steps = [
                (my_step, ['foo', 'bar'],
                 lambda namespace: getattr(namespace, 'foo', False))
                ]

    - add_arguments(parser):

        adds `--steps` and `--skip-steps` arguments to the parser. They can
        be used to include or exclude steps during subcommand's execution.

    - get_description(namespace):

        collects steps' descriptions so that the subcommand description
        is based on the descriptions of its steps.

    - run(namespace):

        implemented to resolve the steps to be run and to invoke them
        in the order they are provided.

    - has_dry_run and has_interactive_run are also set to True because usually
      a steps based subcommand supports interactive and dry run execution.
"""

__metaclass__ = type
__all__ = [
    'add_common_arguments',
    'ArgumentParser',
    'BaseSubCommand',
    'get_args_from_namespace',
    'get_step_description',
    'get_step_name',
    'init_namespace',
    'resolve_steps',
    'restart_as_root',
    'StepsBasedSubCommand',
    ]

import argparse
import os
import string
import subprocess
import sys
import textwrap

from lpsetup.utils import get_terminal_width


class HelpFormatter(argparse.HelpFormatter):
    """A customized help formatter for `argparse`.

    This formatter wraps text, preserving new lines and tabs.
    """

    def _fill_text(self, text, width, indent):
        """Return text wrapped while preserving new lines and tabs."""
        lines = []
        for line in text.splitlines():
            # The default formatter here replaces newlines, multiple spaces
            # and tabs with a single space. Newlines are not present here
            # (since we already called *splitlines()*) and we want to preserve
            # tabs, so we only need to remove multiple whitespaces.
            cleaned = ' '.join(filter(None, line.split(' ')))
            # Wrap the line at given *width*.
            wrapped = textwrap.fill(
                cleaned, width,
                initial_indent=indent, subsequent_indent=indent)
            lines.append(wrapped)
        return '\n'.join(lines)


class ArgumentParser(argparse.ArgumentParser):
    """A customized argument parser for `argparse`."""

    def __init__(self, *args, **kwargs):
        self.actions = []
        super(ArgumentParser, self).__init__(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        """Override to store actions in a "public" instance attribute.

            >>> parser = ArgumentParser()
            >>> parser.add_argument('arg1')
            >>> parser.add_argument('arg2')
            >>> [action.dest for action in parser.actions]
            ['help', 'arg1', 'arg2']
        """
        action = super(ArgumentParser, self).add_argument(*args, **kwargs)
        self.actions.append(action)

    def _get_formatter(self):
        """Override to use a customized help formatter.

        Also create the formatter instance wrapping help at terminal width.
        """
        return HelpFormatter(prog=self.prog, width=get_terminal_width())


class BaseSubCommand(object):
    """Base class defining a subcommand.

    Objects of this class can be used to easily add subcommands to this
    script as plug-ins, providing arguments and validating them.

    See the docstring of this module for further details.
    """

    help = None
    handlers = ()

    def __repr__(self):
        return '<{klass}: {name}>'.format(
            klass=self.__class__.__name__, name=self.help)

    def add_arguments(self, parser):
        """Here subclasses can add subcommand specific args to the parser."""
        pass

    def get_description(self, namespace):
        return ''

    def prepare_namespace(self, namespace):
        """Prepare the current namespace running namespace handlers.

        The method `self.get_handlers` can return an iterable of objects
        that are called once the arguments namespace is fully populated.
        This allows cleaning and validating arguments that depend on
        each other, or on the current environment.

        Each handler is a callable object, takes the current namespace
        and can raise ValidationError if the arguments are not valid.
        """
        for handler in self.get_handlers(namespace):
            handler(namespace)

    def run(self, namespace):
        """Default subcommand callback.

        This method must be implemented by subclasses.
        """
        raise NotImplementedError

    def get_handlers(self, namespace):
        """Return an iterable of namespace handlers for this subcommand.

        Subclasses can override this to dynamically change handlers
        based on the given *namespace*.
        """
        return self.handlers


class StepsBasedSubCommand(BaseSubCommand):
    """A subcommand that uses "steps" to handle its execution.

    Steps are callables stored in the `steps` attribute, together
    with the arguments they expect. Those arguments are strings
    representing attributes of the argparse namespace::

        >>> trace = []

        >>> def step1(foo):
        ...     trace.append('step1 received ' + foo)

        >>> def step2(foo, bar):
        ...     trace.append('step2 received {0} and {1}'.format(foo, bar))

        >>> class SubCommand(StepsBasedSubCommand):
        ...     has_interactive_run = False
        ...     steps = (
        ...         (step1, ['foo']),
        ...         (step2, ['foo', 'bar']),
        ...         )
        ...
        ...     def add_arguments(self, parser):
        ...         super(SubCommand, self).add_arguments(parser)
        ...         parser.add_argument('--foo')
        ...         parser.add_argument('--bar')

    This class implements a `run` method that executes steps in the
    order they are provided.

        >>> subcommand = SubCommand()
        >>> parser = ArgumentParser()
        >>> subcommand.add_arguments(parser)
        >>> namespace = parser.parse_args('--foo eggs --bar spam'.split())
        >>> subcommand.run(namespace)
        >>> trace
        ['step1 received eggs', 'step2 received eggs and spam']

    A special argument `-s` or `--steps` is automatically added to the
    parser. It can be used to execute only one or a subset of steps::

        >>> trace = []

        >>> namespace = parser.parse_args('--foo eggs -s step1'.split())
        >>> subcommand.run(namespace)
        >>> trace
        ['step1 received eggs']

    A special argument `--skip-steps` is automatically added to the
    parser. It can be used to skip one or more steps::

        >>> trace = []

        >>> namespace = parser.parse_args(
        ...     '--foo eggs --skip-steps step1'.split())
        >>> subcommand.run(namespace)
        >>> trace
        ['step2 received eggs and None']
    """

    steps = ()
    has_dry_run = True
    has_interactive_run = True

    def add_arguments(self, parser):
        """Add steps related arguments."""
        step_names = [get_step_name(i[0]) for i in self.steps]
        parser.add_argument(
            '-s', '--steps',
            choices=step_names,  metavar='STEP_NAME', nargs='+',
            # argparse internally formats the help string using choices
            # if they are present as an attribute of the action.
            help='Call one or more internal functions. '
                 'Available steps for this subcommand, in the order they '
                 'are executed: %(choices)s.')
        parser.add_argument(
            '--skip-steps',
            choices=step_names, metavar='STEP_NAME', nargs='+',
            help='Skip one or more internal functions. See the help above '
                 'for --steps for the list of available steps.')

    def get_description(self, namespace):
        """Collect steps' descriptions and return them as a single string."""
        context = namespace.__dict__
        steps = resolve_steps(self.steps, namespace)
        descriptions = [get_step_description(i[0], **context) for i in steps]
        return '\n'.join(filter(None, descriptions))

    def run(self, namespace):
        """Execute the steps in the order they are provided."""
        for step, args in resolve_steps(self.steps, namespace):
            step(*args)


def add_common_arguments(
    parser, has_dry_run=False, has_interactive_run=False):
    """Add to given *parser* the arguments that all subcommands have in common.

    The common arguments are:

        - `-y` or `--yes`: non-interactive execution.
        - `--dry-run`: dry run execution.

    Ensure the resulting namespace always contains *interactive* and *dry_run*
    names, even if *has_interactive_run* or *has_dry_run* are False.
    """
    if has_dry_run:
        parser.add_argument('--dry-run', action='store_true', help='Dry run.')
    else:
        parser.set_defaults(dry_run=False)
    if has_interactive_run:
        parser.add_argument(
            '-y', '--yes', action='store_false', dest='interactive',
            help='Assume yes to all queries.')
    else:
        parser.set_defaults(interactive=False)


def get_args_from_namespace(parser, namespace):
    """Return a list of arguments taking values from `namespace`."""
    args = []
    for action in parser.actions:
        dest = action.dest
        option_strings = action.option_strings
        value = getattr(namespace, dest, None)
        isbool = isinstance(value, bool)
        # If the value is boolean and the action is 'store_false', we
        # invert the value. This way the following `if value:` block
        # is executed if the original value is False, and the argument
        # is correctly added.
        if isbool and isinstance(action, argparse._StoreFalseAction):
            value = not value
        if value:
            if option_strings:
                args.append(option_strings[0])
            if isinstance(value, list):
                args.extend(value)
            elif not isbool:
                args.append(value)
    return args


def init_namespace(namespace):
    """Add `run_as_root` and `euid` names to the given *namespace*."""
    euid = os.geteuid()
    namespace.euid, namespace.run_as_root = euid, not euid


def restart_as_root(parser, namespace):
    """Restart this script using *sudo*.

    The arguments are recreated using the given `namespace`.
    """
    cmd = ['sudo', sys.argv[0]] + parser.prog.split()[1:]
    namespace.interactive = False
    args = get_args_from_namespace(parser, namespace)
    return subprocess.call(cmd + args)


def get_step_name(step):
    """Return a string representing the name of the given *step*."""
    try:
        return step.step_name
    except AttributeError:
        return step.__name__


def resolve_steps(steps, namespace):
    """Return a list of *(step_callable, step_args)* tuples.

    The argument *steps* is a sequence of
    *(step_callable, arg_names, filter_callable)*.

    The *arg_names* are strings representing arguments in the namespace.
    The *filter_callable* is optional. It must take the namespace and return
    True if the corresponding step must be included, False otherwise.

    If the namespace contains the `steps` or `skip_steps` names, they are
    taken into consideration to resolve the steps that must be run.

    In essence, a step is returned as part of the sequence if:

        - is not included in *namespace.skip_steps*
        - is not excluded from *namespace.step* (if *namespace.step* is not
          defined or is None, all the steps are considered as included)
        - the provided filter callable returns True (if not present, the
          step is included by default)
    """
    steps_to_skip = getattr(namespace, 'skip_steps', None) or []
    steps_to_run = getattr(namespace, 'steps', None)

    def get_step(step_sequence):
        if len(step_sequence) == 3:
            # If filter_callable is present, just return the given sequence.
            return step_sequence
        # Otherwise, add a dummy filter_callable that always returns True.
        return list(step_sequence) + [lambda namespace: True]

    step_args = []
    for step_callable, arg_names, should_run_step in map(get_step, steps):
        step_name = get_step_name(step_callable)
        # Include step if is not explicitly told to skip it.
        if step_name not in steps_to_skip:
            # If no list of steps was provided then any non-skipped are to be
            # run, otherwise the step has to be included in order to be run.
            if (steps_to_run is None) or (step_name in steps_to_run):
                # The step filter function, if present, can deny the execution.
                if should_run_step(namespace):
                    args = [getattr(namespace, i) for i in arg_names]
                    step_args.append((step_callable, args))
    return step_args


def get_step_description(step, **kwargs):
    """Retrieve and format step description from the given *step* callable.

    *kwargs*, if provided, will be used as context to format the description.
    Formatting is done using the Python's built-in templating system
    *string.Template* supporting $-based substitutions.

    If placeholders are missing from *kwargs* an error will be raised.
    """
    description = step.description
    if not description:
        # This step has no description, nothing else to do.
        return ''
    if kwargs:
        s = string.Template(description)
        description = s.substitute(**kwargs)
    # Remove multiple spaces from lines.
    lines = [' '.join(line.split()) for line in description.splitlines()]
    # Retrieve all the non empty lines.
    lines = filter(None, lines)
    # For each line, wrap the contents. Note that we can't wrap the text of
    # the entire paragraph because we want to preserve existing new lines.
    width = get_terminal_width()
    return '\n'.join(textwrap.fill(
        line, width=width, initial_indent='* ', subsequent_indent='  ')
        for line in lines)
