#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Subcommand arguments handling and validation."""

__metaclass__ = type
__all__ = [
    'handle_branch_and_checkout',
    'handle_directories',
    'handle_lpuser_as_username',
    'handle_lpuser_from_lplogin',
    'handle_source',
    'handle_ssh_keys',
    'handle_target_dir',
    'handle_target_from_repository',
    'handle_testing',
    'handle_user',
    'handle_userdata',
    ]

import os
import pwd
import subprocess

from shelltoolbox import (
    bzr_whois,
    environ,
    get_user_home,
    run,
    user_exists,
    )

from lpsetup.exceptions import ValidationError
from lpsetup.settings import (
    LP_HTTP_REPO,
    LP_SSH_REPO,
    )


def handle_user(namespace):
    """Handle user argument.

    This handler populates namespace with `home_dir` name::

        >>> import getpass
        >>> username = getpass.getuser()

        >>> import argparse
        >>> namespace = argparse.Namespace(user=username)

        >>> handle_user(namespace)
        >>> namespace.home_dir == '/home/' + username
        True

    The validation fails if the current user is root and no user is provided::

        >>> namespace = argparse.Namespace(user=None, euid=0)
        >>> handle_user(namespace) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: argument user ...
    """
    if getattr(namespace, 'user', None) is None:
        if not namespace.euid:
            raise ValidationError('argument user can not be omitted if '
                                  'the script is run as root.')
        namespace.user = pwd.getpwuid(namespace.euid).pw_name
    namespace.home_dir = get_user_home(namespace.user)


def handle_lpuser_as_username(namespace):
    """Handle lpuser argument.

    The namespace must contain *user* and *lpuser*.
    It should be invoked after handle_user.

    If lpuser is None, the user name is used::

        >>> import getpass
        >>> username = getpass.getuser()

        >>> import argparse
        >>> namespace = argparse.Namespace(user=username, lpuser=None)
        >>> handle_lpuser_as_username(namespace)
        >>> namespace.lpuser == username
        True
    """
    if namespace.lpuser is None:
        namespace.lpuser = namespace.user


def handle_lpuser_from_lplogin(namespace):
    """Handle lpuser argument.

    The validation fails if *lpuser* is not provided, is not retrievable
    and *use-http* is False.
    """
    if getattr(namespace, 'lpuser', None) is None:
        try:
            namespace.lpuser = run('bzr', 'launchpad-login').strip()
        except subprocess.CalledProcessError:
            if getattr(namespace, 'use_http', False):
                namespace.lpuser = None
            else:
                raise ValidationError("No bzr launchpad-login set.")


def handle_userdata(namespace, whois=bzr_whois):
    """Handle full_name and email arguments.

    If they are not provided, this function tries to obtain them using
    the given `whois` callable::

        >>> import argparse
        >>> namespace = argparse.Namespace(
        ...     full_name=None, email=None, user='root')
        >>> email = 'email@example.com'
        >>> handle_userdata(namespace, lambda user: (user, email))
        >>> namespace.full_name == namespace.user
        True
        >>> namespace.email == email
        True

    The validation fails if full_name or email are not provided and
    they can not be obtained using the `whois` callable::

        >>> namespace = argparse.Namespace(
        ...     full_name=None, email=None, user='root')
        >>> handle_userdata(namespace, lambda user: None) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: arguments full-name ...

        >>> namespace = argparse.Namespace(
        ...     full_name=None, email=None, user='this_user_does_not_exist')
        >>> handle_userdata(namespace) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: arguments full-name ...

    It does not make sense to provide only one argument::

        >>> namespace = argparse.Namespace(full_name='Foo Bar', email=None)
        >>> handle_userdata(namespace) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: arguments full-name ...
    """
    args = (namespace.full_name, namespace.email)
    if not all(args):
        if any(args):
            raise ValidationError(
                'arguments full-name and email: '
                'either none or both must be provided.')
        if user_exists(namespace.user):
            userdata = whois(namespace.user)
            if userdata is None:
                raise ValidationError(
                    'arguments full-name and email are required: '
                    'bzr user id not found.')
            namespace.full_name, namespace.email = userdata
        else:
            raise ValidationError(
                'arguments full-name and email are required: '
                'system user not found.')


def handle_ssh_keys(namespace):
    r"""Handle private and public ssh keys.

    The namespace must contain *home_dir* and *ssh_key_name*.
    It should be invoked after handle_user.

    This handler adds *ssh_key_path* and *valid_ssh_keys* to the namespace.
    """
    namespace.ssh_key_path = os.path.join(
        namespace.home_dir, '.ssh', namespace.ssh_key_name)
    namespace.valid_ssh_keys = os.path.isfile(namespace.ssh_key_path)
    public_key_path = namespace.ssh_key_path + '.pub'
    # Exits with an error if only one key of the ssh pair is found.
    if namespace.valid_ssh_keys != os.path.isfile(public_key_path):
        raise ValidationError(
            'ssh private and public keys {0}(.pub): both must exist '
            'or neither must exist.'.format(namespace.ssh_key_path))


def handle_directories(namespace):
    """Handle repository and dependencies directories.

    The namespace must contain *home_dir*.
    It should be invoked after handle_user.

    - The ~ construction is automatically expanded.
    - The validation fails for directories not residing inside the home.
    - The validation fails if the directory contains spaces.
    """
    for attr in ('repository', 'dependencies_dir'):
        directory = getattr(namespace, attr, None)
        if directory is None:
            continue
        if ' ' in directory:
            err = ("argument directory cannot contain "
                   "spaces: '{0}'".format(directory))
            raise ValidationError(err)
        # To expand the "~" construction, *os.path.expanduser* tries
        # to use the "HOME" environment variable.
        # Using *shelltoolbox.environ*, here we replace the original env var
        # with the home dir already present in the namespace, for 2 reasons:
        # - it is possible that we don't need the current user's home
        #   (e.g. lpsetup is run as root)
        # - it is possible that we need the home of a non-existent user
        #   (e.g. lpsetup needs to create the user specified using `--user`)
        with environ(HOME=namespace.home_dir):
            directory = os.path.expanduser(directory)
        directory = os.path.abspath(directory)
        if not directory.startswith(namespace.home_dir + os.path.sep):
            raise ValidationError(
                'argument {0} does not reside under the home '
                'directory of the system user.'.format(attr))
        setattr(namespace, attr, directory)


def handle_testing(namespace):
    """Handle the `testing` flag.

    Set the parallel testing related options to True if `testing` is True.
    """
    if getattr(namespace, 'testing', False):
        namespace.create_scripts = True
        namespace.install_haveged = True
        namespace.install_subunit = True
        namespace.interactive = False
        namespace.no_checkout = True
        namespace.stop_lxc = True
        namespace.use_http = True


def handle_source(namespace):
    """Handle the --source and --use-http options.

    The namespace must contain *source* or *use_http*.
    """
    # If a source has been provided, then there is nothing else to do.
    if namespace.source is not None:
        return
    if namespace.use_http:
        namespace.source = LP_HTTP_REPO
    else:
        namespace.source = LP_SSH_REPO


def handle_target_from_repository(namespace):
    """Handle *repository*, *branch_name*, *checkout_name* and *no_checkout*.

    The namespace must contain *repository*, *branch_name*,
    *checkout_name* and *no_checkout*.
    It should be invoked after handle_directories, and when *target_dir* is
    not explicitly passed to a command.

    These names must be present in the namespace.
    Produces *target_dir*: the real working dir where the actual code lives.
    """
    if namespace.no_checkout:
        target_name = namespace.branch_name
    else:
        target_name = namespace.checkout_name
    namespace.target_dir = os.path.join(namespace.repository, target_name)


def handle_target_dir(namespace):
    """Handle path to the working directory.

    The namespace must contain *target_dir*.
    Raise *ValidationError* if the resulting target dir:

        - does not exist
        - is not a directory
        - is not writable by the current user
        - does not seem a valid Launchpad branch

    It should be invoked when *target_dir* is explicitly passed to a command.
    """
    target_dir = os.path.abspath(os.path.expanduser(namespace.target_dir))
    # The target dir must be an existing directory.
    if not os.path.isdir(target_dir):
        raise ValidationError(
            'the target dir {0} does not exist or is not a directory.'.format(
                target_dir))
    # The target dir must be writable.
    if not os.access(target_dir, os.W_OK):
        raise ValidationError(
            'the target dir {0} is not writable by the current user.'.format(
                target_dir))
    # The target dir must look like a Launchpad tree.
    # This check is done in a naive way: `utilities/update-sourcecode`
    # must be included in the target dir.
    path_to_check = os.path.join(target_dir, 'utilities', 'update-sourcecode')
    if not os.path.exists(path_to_check):
        raise ValidationError(
            'the target dir {0} is not a valid Launchpad tree.'.format(
                target_dir))
    namespace.target_dir = target_dir


def handle_branch_and_checkout(namespace):
    """Handle branch and checkout names.

    Raise *ValidationError* if::

    - branch/checkout name is '.' or '..'
    - branch/checkout name contains '/'
    - lightweight checkout is used and branch and checkout have the same name

    The namespace must contain the following names::

        - branch_name
        - checkout_name
        - repository

    This handler does not modify the namespace.
    """
    branch_name = namespace.branch_name
    checkout_name = namespace.checkout_name
    for name in (branch_name, checkout_name):
        if name in ('.', '..') or '/' in name:
            raise ValidationError(
                'invalid name "{0}": branch or checkout names can not '
                'be "." or ".." and can not contain "/"'.format(name))
    is_lightweight = not getattr(namespace, 'no_checkout', False)
    if is_lightweight and (checkout_name == branch_name):
        raise ValidationError(
            'branch and checkout: can not use the same name ({0}).'.format(
                checkout_name))
