#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""init-lxc subcommand: prepare an LXC container to run Launchpad.

Run as root on the host machine to create an LXC container that can later be
used to install a Launchpad development environment.
"""

__metaclass__ = type
__all__ = [
    'create_lxc',
    'init_target_in_lxc',
    'initialize',
    'install_lpsetup_in_lxc',
    'start_lxc',
    'stop_lxc',
    'SubCommand',
    'wait_for_lxc',
    ]

import os
import shutil
import subprocess
import tempfile

from shelltoolbox import (
    apt_get_install,
    file_append,
    get_su_command,
    mkdirs,
    su,
    )

from lpsetup import exceptions
from lpsetup.settings import (
    LPSETUP_PPA,
    LXC_CONFIG_TEMPLATE,
    LXC_GUEST_ARCH,
    LXC_GUEST_CHOICES,
    LXC_GUEST_OS,
    LXC_NAME,
    LXC_OPTIONS,
    LXC_PACKAGES,
    PY26_PACKAGES,
    )
from lpsetup.subcommands import init_target
from lpsetup.utils import (
    call,
    get_container_path,
    get_lxc_gateway,
    lxc_stopped,
    retry,
    sshlxc as ssh,
    this_command,
    wrap_contents,
    )


def initialize(user, install_haveged):
    """Initialize the LXC host."""
    init_target.initialize_base(user)
    if install_haveged:
        # haveged is used to fill /dev/random, avoiding
        # entropy exhaustion during automated parallel tests.
        apt_get_install('haveged', caller=call)

initialize.description = init_target.initialize_base.description + """If \
    requested, install haveged in order to fill /dev/random, avoiding \
    entropy exhaustion during automated parallel tests.
"""


def create_lxc(lxc_name, lxc_arch, lxc_os, user, install_subunit):
    """Create the LXC named `lxc_name` sharing `user` home directory.

    The container will be used as development environment or as base template
    for parallel testing using ephemeral instances.
    """
    # Install necessary deb packages.
    packages = list(LXC_PACKAGES)
    if install_subunit:
        packages.append('subunit')
    apt_get_install(*packages, caller=call)
    # XXX 2012-02-02 gmb bug=925024:
    #     These calls need to be removed once the lxc vs. apparmor bug
    #     is resolved, since having apparmor enabled for lxc is very
    #     much a Good Thing.
    # Disable the apparmor profiles for lxc so that we don't have
    # problems installing postgres.
    subprocess.call([
        'ln', '-s',
        '/etc/apparmor.d/usr.bin.lxc-start', '/etc/apparmor.d/disable/'])
    subprocess.call([
        'apparmor_parser', '-R', '/etc/apparmor.d/usr.bin.lxc-start'])
    # Container configuration template.
    lxc_gateway = get_lxc_gateway()
    if lxc_gateway is None:
        raise exceptions.ExecutionError(
            'Error: LXC bridge interface not found.')
    contents = wrap_contents(LXC_OPTIONS.format(interface=lxc_gateway))
    with open(LXC_CONFIG_TEMPLATE, 'w') as f:
        f.write(contents)
    # Creating container.
    call(
        'lxc-create',
        '-t', 'ubuntu',
        '-n', lxc_name,
        '-f', LXC_CONFIG_TEMPLATE,
        '--',
        '-r {os} -a {arch} -b {user}'.format(
            os=lxc_os, arch=lxc_arch, user=user),
        )
    # Set up root ssh key.
    user_authorized_keys = os.path.expanduser(
        '~' + user + '/.ssh/authorized_keys')
    dst = get_container_path(lxc_name, '/root/.ssh/')
    mkdirs(dst)
    shutil.copy(user_authorized_keys, dst)
    if lxc_os == 'lucid':
        # XXX 2012-04-18 frankban bug=974584:
        # Add a line to the container's fstab to be able to create semaphores
        # in lxc. This workaround needs to be removed once the lxc bug is
        # resolved for lucid containers too.
        file_append(
            '/var/lib/lxc/{0}/fstab'.format(lxc_name),
            'none dev/shm tmpfs defaults 0 0\n')

create_lxc.description = """Create an LXC container named $lxc_name \
    ($lxc_os $lxc_arch), bind mounting $home_dir, and disabling apparmor \
    profiles for lxc so that we don't have problems installing Postgres.
    Allow root access to the container.
"""


def start_lxc(lxc_name):
    """Start the lxc instance named `lxc_name`."""
    call('lxc-start', '-n', lxc_name, '-d')

start_lxc.description = 'Start the LXC container named $lxc_name.\n'


def wait_for_lxc(lxc_name, ssh_key_path):
    """Try to ssh into the LXC container named `lxc_name`."""
    retry_ssh = retry(subprocess.CalledProcessError)(ssh)
    retry_ssh(lxc_name, 'true', key=ssh_key_path)

wait_for_lxc.description = None


def install_lpsetup_in_lxc(lxc_name, ssh_key_path, lxc_os,
                           user, home_dir, lpsetup_branch=None):
    """Initialize LXC container."""
    # Install python-software-properties in order to get 'apt-add-repository',
    # which is a pre-requisite to installing our PPA.
    # http://tinyurl.com/python-software-properties
    # Do it even if we're installing from a branch because the repository may
    # be needed in the future.
    ssh(lxc_name,
        'DEBIAN_FRONTEND=noninteractive '
        'apt-get install -y python-software-properties',
        key=ssh_key_path)
    args = {
        'assume_yes': '' if lxc_os == 'lucid' else '-y',
        'repository': LPSETUP_PPA,
        }
    ssh(lxc_name,
        'apt-add-repository {assume_yes} {repository}'.format(**args),
        key=ssh_key_path)
    ssh(lxc_name, 'apt-get clean && apt-get update', key=ssh_key_path)
    # Install lpsetup.
    if lpsetup_branch is None:
        ssh(lxc_name,
            'DEBIAN_FRONTEND=noninteractive '
            'apt-get upgrade -y && apt-get install -y lpsetup',
            key=ssh_key_path)
    else:
        # Make a temporary directory under the user's shared home directory to
        # hold the lpsetup branch.
        with su(user):
            # Using mktemp here despite the warnings since we just need a file
            # name in the user's home directory.
            dest = tempfile.mktemp(dir=home_dir)
            # Bzr requires the use of get_su_command due to the way it
            # handles uids.
            call(*get_su_command(user, [
                'bzr', 'checkout', '--lightweight', lpsetup_branch, dest]))
        if lxc_os == 'lucid':
            # If the container is lucid we must manually install a few
            # packages.  These packages are dependencies of the lpsetup
            # package and would normally be installed with the lpsetup
            # package.
            ssh(lxc_name,
                'DEBIAN_FRONTEND=noninteractive '
                'apt-get install -y {}'.format(' '.join(PY26_PACKAGES)),
                key=ssh_key_path)

        ssh(lxc_name,
            'DEBIAN_FRONTEND=noninteractive '
            'cd {} && python setup.py install'.format(dest),
            key=ssh_key_path)
        # Remove the lpsetup tree from the user's directory on the host.  This
        # must be done as root.
        shutil.rmtree(dest)

install_lpsetup_in_lxc.description = None


def init_target_in_lxc(lxc_name, ssh_key_path, user, email, full_name, lpuser,
                    ssh_key_name, home_dir):
    """Prepare the Launchpad environment inside an LXC."""
    # Use ssh to call this script from inside the container.
    args = ['init-target', '--yes', '-u', user, '-E', email, '-f', full_name,
            '-l', lpuser, '-S', ssh_key_name, '--skip-steps', 'setup_home']
    cmd = this_command(home_dir, args)
    ssh(lxc_name, cmd, key=ssh_key_path)

init_target_in_lxc.description = 'Initialize the LXC instance $lxc_name.\n'


def stop_lxc(lxc_name, ssh_key_path):
    """Stop the lxc instance named `lxc_name`."""
    ssh(lxc_name, 'poweroff', key=ssh_key_path)
    if not lxc_stopped(lxc_name):
        subprocess.call(['lxc-stop', '-n', lxc_name])

stop_lxc.description = 'Stop the LXC instance $lxc_name.\n'


class SubCommand(init_target.SubCommand):
    """Create an LXC container suitable for later installing a Launchpad
    development environment.
    """

    create_lxc_step = (
        create_lxc,
        ['lxc_name', 'lxc_arch', 'lxc_os', 'user', 'install_subunit'])
    start_lxc_step = (start_lxc, ['lxc_name'])
    wait_for_lxc_step = (wait_for_lxc, ['lxc_name', 'ssh_key_path'])
    install_lpsetup_in_lxc_step = (
        install_lpsetup_in_lxc,
        ['lxc_name', 'ssh_key_path', 'lxc_os', 'user', 'home_dir',
            'lpsetup_branch'])
    init_target_in_lxc_step = (
        init_target_in_lxc,
        ['lxc_name', 'ssh_key_path', 'user', 'email', 'full_name', 'lpuser',
            'ssh_key_name', 'home_dir'])
    stop_lxc_step = (
        stop_lxc, ['lxc_name', 'ssh_key_path'],
        lambda namespace: namespace.stop_lxc)

    base_steps = (
        (initialize, ['user', 'install_haveged']),
        init_target.SubCommand.setup_home_step,
        create_lxc_step,
        start_lxc_step,
        wait_for_lxc_step,
        install_lpsetup_in_lxc_step,
        init_target_in_lxc_step,
        )
    steps = base_steps + (stop_lxc_step, )

    help = __doc__
    root_required = True

    epilog = """`%(prog)s` creates an LXC container bind mounting the home \
        directory of the current user, setting up ssh keys if needed and \
        initializing Launchpad dependencies.

        Create and initialize an LXC named `mylxc`:

        \t$ %(prog)s mylxc

        The default LXC name is `{lxc_name}`, so, to create and initialize \
        an LXC named `{lxc_name}`, you can avoid passing the name:

        \t$ %(prog)s

        By default the home of the current user is bind mounted, but if this \
        subcommand is run as root, you need to explicitly specify the user:

        \t# %(prog)s -u myuser

        In this case, if `myuser` does not exist, it will be created.

        The current user (or the provided one) should have a correctly \
        configured bazaar user id (i.e `bzr whoami`). However, sometimes this \
        is not possible (e.g. when the provided user does not exist). In this \
        case, you can explicitly pass the user's full name and email and \
        the bazaar user id will be set up using provided arguments:

        \t# %(prog)s -u myuser -f 'My User' -e myuser@example.com

        At the end of the process the newly created LXC instance is running, \
        and it is possible to obtain its ip address using:

        \t$ sudo lp-lxc-ip -n mylxc

        If instead you want `%(prog)s` to stop the instance once it is \
        completely configured, you can use the `--stop-lxc` argument:

        \t$ %(prog)s --stop-lxc
    """.format(lxc_name=LXC_NAME)

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        # Add LXC related arguments.
        parser.add_argument(
            'lxc_name', nargs='?', default=LXC_NAME,
            help='The LXC container name to setup. [DEFAULT=%(default)s]')
        parser.add_argument(
            '-A', '--lxc-arch', default=LXC_GUEST_ARCH,
            help='The LXC container architecture. [DEFAULT=%(default)s]')
        parser.add_argument(
            '-R', '--lxc-os', default=LXC_GUEST_OS,
            choices=LXC_GUEST_CHOICES,
            help='The LXC container distro codename. [DEFAULT=%(default)s]')
        parser.add_argument(
            '--stop-lxc', action='store_true',
            help='Shut down the LXC container at the end of the process.')
        parser.add_argument(
            '--omit-subunit', action='store_false', dest='install_subunit',
            help='Do not install subunit in the host machine. Subunit should '
                 'be usually installed if you are using this command to '
                 'create an automated parallel testing environment.')
        parser.add_argument(
            '--omit-haveged', action='store_false', dest='install_haveged',
            help='Do not install haveged in the host machine. Haveged is '
                 'used to fill /dev/random, avoiding entropy exhaustion '
                 'during automated parallel tests.')
        parser.add_argument(
            '-B', '--lpsetup-branch', default=None,
            help='The Launchpad branch to use for lpsetup inside the '
                 'container.  Useful for testing lpsetup changes.  '
                 'Leave unset to install from a Debian package.')
