#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""LXC install subcommand: create a Launchpad environment inside an LXC."""

__metaclass__ = type
__all__ = [
    'create_scripts',
    'init_repo_in_lxc',
    'SubCommand',
    'update_in_lxc',
    ]

import os

from lpsetup.handlers import (
    handle_branch_and_checkout,
    handle_directories,
    handle_source,
    handle_target_from_repository,
    handle_testing,
    )
from lpsetup.settings import (
    LXC_IP_COMMAND,
    TARGET_SCRIPTS,
    )
from lpsetup.subcommands import (
    initlxc,
    initrepo,
    update,
    )
from lpsetup.utils import (
    get_file_header,
    render_to_file,
    sshlxc as ssh,
    this_command,
    wrap_contents,
    )


def create_script(template_name, dest, context):
    """Create an individual script from a template."""
    render_to_file(template_name, context, dest)
    os.chmod(dest, 0555)
    return dest


def create_scripts(lxc_name, ssh_key_path, user):
    """Create scripts to update the Launchpad environment and run tests.

    These scripts are only needed to support the Buildbot testing scenario and
    thus are only needed in the container.
    """
    context = {
        'lxcip': ' '.join(LXC_IP_COMMAND),
        'lxc_name': lxc_name,
        'ssh_key_path': ssh_key_path,
        'user': user,
        'installation_header': get_file_header(initial_indent=''),
        }
    scripts = []
    # Generate the scripts.
    for template_name in TARGET_SCRIPTS:
        dest = os.path.join(os.path.sep, 'usr', 'local', 'bin', template_name)
        scripts.append(create_script(template_name, dest, context))
    # Add a file to sudoers.d that will let the buildbot user run the
    # above.
    sudoers_file = '/etc/sudoers.d/launchpad-' + user
    sudoers_contents = '{user} ALL = (ALL) NOPASSWD: {scripts}\n'.format(
        user=user, scripts=', '. join(scripts))
    with open(sudoers_file, 'w') as sudoers:
        sudoers.write(wrap_contents(sudoers_contents))
    # The sudoers must have this mode or it will be ignored.
    os.chmod(sudoers_file, 0440)
    # XXX 2012-03-13 frankban bug=944386:
    #     Disable hardlink restriction. This workaround needs
    #     to be removed once the kernel bug is resolved.
    procfile = '/proc/sys/kernel/yama/protected_nonaccess_hardlinks'
    with open(procfile, 'w') as f:
        f.write('0\n')


create_scripts.description = """Create helper scripts \
    /usr/local/bin/lp-setup-*: they can be used to build Launchpad and \
    start a parallel test run.
"""


def cmd_in_lxc(lxc_name, ssh_key_path, home_dir, args, as_user=None):
    cmd = this_command(home_dir, args)
    ssh(lxc_name, cmd, key=ssh_key_path, user=as_user)


def init_repo_in_lxc(
    lxc_name, ssh_key_path, home_dir, user, source, use_http,
    branch_name, checkout_name, repository, no_checkout):
    args = [
        'init-repo', '--source', source,
        '--branch-name', branch_name, '--checkout-name', checkout_name,
        '--repository', repository, '--yes',
        ]
    if use_http:
        args.append('--use-http')
    if no_checkout:
        args.append('--no-checkout')
    cmd_in_lxc(lxc_name, ssh_key_path, home_dir, args, as_user=user)

init_repo_in_lxc.description = """Initialize the Launchpad repository \
    inside the LXC container $lxc_name.
""" + initrepo.fetch.description + initrepo.setup_bzr_locations.description


def update_in_lxc(
    lxc_name, ssh_key_path, home_dir, user, external_path,
    target_dir, lp_source_deps, use_http):
    args = ['update', target_dir, '--external-path', external_path,
            '--lp-source-deps', lp_source_deps]
    if use_http:
        args.append('--use-http')
    cmd_in_lxc(lxc_name, ssh_key_path, home_dir, args, as_user=user)

update_in_lxc.description = (
    update.initialize_directories.description +
    update.update_dependencies.description +
    update.update_tree.description)


def finish_install_in_lxc(
    lxc_name, ssh_key_path, user, target_dir):
    args = ['install']
    cmd = this_command(target_dir, args, cmd='make')
    ssh(lxc_name, cmd, key=ssh_key_path)

    # Change owner of /srv/launchpad.dev/ by running the following in the
    # container as root:
    # % chown user:user /srv/launchpad.dev
    args = ['{0}:{0}'.format(user), '/srv/launchpad.dev']
    cmd = this_command('/', args, cmd='chown')
    ssh(lxc_name, cmd, key=ssh_key_path)

finish_install_in_lxc.description = """Install Launchpad and \
    configure Apache.

    Also change ownership of /srv/launchpad.dev to the user.
"""


class SubCommand(initlxc.SubCommand):
    """Completely sets up an LXC environment with Launchpad."""

    steps = initlxc.SubCommand.base_steps + (
        # Run on host:
        (create_scripts, ['lxc_name', 'ssh_key_path', 'user'],
         lambda namespace: namespace.create_scripts),
        # Run inside the container:
        (init_repo_in_lxc,
         ['lxc_name', 'ssh_key_path', 'home_dir', 'user', 'source', 'use_http',
          'branch_name', 'checkout_name', 'repository', 'no_checkout']),
        (update_in_lxc,
         ['lxc_name', 'ssh_key_path', 'home_dir', 'user', 'external_path',
          'target_dir', 'lp_source_deps', 'use_http']),
        (finish_install_in_lxc,
         ['lxc_name', 'ssh_key_path', 'user', 'target_dir']),
        # Run on host:
        initlxc.SubCommand.stop_lxc_step,
        )
    help = __doc__

    epilog = """`%(prog)s` sets up a complete Launchpad environment in an LXC \
    container, creating the LXC, configuring the instance, retrieving the code.

    This is basically the same as running `init-lxc` and then, inside the \
    newly created container, `lp-setup init-repo`, `lp-setup update`, and
    `make install`.

    Set up Launchpad for the current user:

    \t%(prog)s

    A more complex example: set up an environment for `myuser`, creating a \
    repository in a customized path, using a different ssh key and activating \
    the parallel testing tweaks.

    \t%(prog)s -u myuser -E myuser@example.com -f 'My User' \
    -r ~/my-lp-branches -S launchpad_lxc_id_rsa --testing

    %(prog)s inherits its arguments from the subcommands it wraps. \
    You can refer to the help of all the other subcommands for a more \
    exhaustive explanation.
    """

    def get_handlers(self, namespace):
        handlers = super(SubCommand, self).get_handlers(namespace)
        return handlers + (
            handle_testing,
            handle_directories,
            handle_branch_and_checkout,
            handle_source,
            handle_target_from_repository,
            )

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        # Inherit arguments from subcommands we depend upon.
        initrepo.SubCommand.add_common_arguments(parser)
        update.SubCommand.add_common_arguments(parser)
        # Add parallel testing related arguments.
        parser.add_argument(
            '-C', '--create-scripts', action='store_true',
            help='Create the scripts used by buildbot for parallel testing.')
        parser.add_argument(
            '--create-lxc-scripts', action='store_true',
            help='Create the required scripts in the container.')
        parser.add_argument(
            '--testing', action='store_true',
            help='Alias for --create-scripts --no-checkout --stop_lxc '
                 '--use-http --yes. This flag also forces subunit and '
                 'haveged installation in the host machine.')
