#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""A simple, end-to-end integration test of install-lxc."""

import sys
import urlparse

from bzrlib.branch import Branch

from common import IntegrationTestBase


class InstallLXCTest(IntegrationTestBase):

    test_type = 'local LXC target'
    default_juju_env = 'lpsetup-testing-lxc'
    juju_target = 'ubuntu/0'
    test_type = 'LXC container'
    repo = '~/launchpad-testing/lp-branches'

    def get_branch(self, dir='.'):
        branch, _ = Branch.open_containing(dir)
        return branch

    def get_remote_url(self, branch=None):
        if branch is None:
            branch = self.get_branch()
        url = (branch.get_push_location() or
               branch.get_parent() or
               branch.get_submit_branch() or
               branch.get_bound_location())
        if url is not None and url.startswith('lp:'):
            # Convert to bzr+ssh or the calls to WorkingTree fail,
            # inexplicably.
            url = url.replace('lp:', 'bzr+ssh://bazaar.launchpad.net/')
        return url

    def check_environment(self):
        """Be sure the test environment doesn't exist."""

        # We want to be really sure we do not clobber an already-existing juju
        # environment.  Therefore we make sure one doesn't exist before
        # bootstrapping.
        super(InstallLXCTest, self).check_environment()

        # Ensure the local branch has been pushed.
        # For some reason 'bzr missing' cannot use a 'lp:' URL.  http or
        # bzr+ssh URLs work.
        self.branch = self.get_branch()
        self.remote_url = self.get_remote_url(self.branch)
        if self.remote_url is None:
            # If no remote_url is found then just skip the rest of the checks
            # as it may be a run inside the tarmac environment.
            # This approach is less than ideal but required by tarmac.
            raise RuntimeError(
                'A branch for this code cannot be found on Launchpad, which '
                'is required for the test to run.')

    def set_up(self):
        super(InstallLXCTest, self).set_up()
        self.on_remote('cd lpsetup; sudo python setup.py install')
        # The installation of `lxc` is redundant as init-lxc does it but
        # required to get the nested LXC work-around to work.
        self.on_remote('sudo apt-get -y install lxc')
        self.on_remote('sudo stop lxc')
        self.on_remote('sudo sed -i "s/10.0.3/10.0.4/g" /etc/default/lxc')
        self.on_remote('sudo start lxc')

    def do_test(self):
        """Run an end-to-end integration tests of the LXC lpsetup story."""
        # Since the most common scenario is to have bzr whoami setup, we do
        # that instead of providing the arguments directly to lpsetup.
        super(InstallLXCTest, self).do_test()
        urlparts = list(urlparse.urlsplit(self.remote_url))
        urlparts[0] = 'http'
        branch_location = urlparse.urlunsplit(urlparts)
        lp_source_stub = 'http://bazaar.launchpad.net/~yellow/launchpad/stub'
        lp_source_deps_stub = (
            'http://bazaar.launchpad.net/~yellow/lp-source-dependencies/stub')
        cmd = (
            'lpsetup/lp-setup install-lxc --use-http '
            '--source {} -B {} -r {} --yes '
            '--lp-source-deps {}'.format(
            lp_source_stub, branch_location, self.repo, lp_source_deps_stub))
        self.on_remote(cmd)


if __name__ == '__main__':
    sys.exit(InstallLXCTest().run())
