#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the cli module."""

from contextlib import nested
import unittest

from lpsetup import (
    cli,
    exceptions,
    )
from lpsetup.tests.utils import (
    call_replaced_by,
    capture_error,
    capture_output,
    ParserTestMixin,
    RawInputReturning,
    SubCommandFactoryTestMixin,
    )


class GetParserTest(SubCommandFactoryTestMixin, unittest.TestCase):

    def test_base_help(self):
        # Ensure the help subcommand is correctly added to the parser.
        parser = cli.get_parser([])
        namespace = parser.parse_args(['help'])
        with self.assertRaises(SystemExit) as cm:
            with capture_output() as output:
                namespace.main(namespace)
        self.assertEqual(0, cm.exception.code)
        self.assertIn('usage:', output.getvalue())

    def test_subcommand_in_global_help(self):
        # The help attribute of subcommands is used to generate
        # the base command usage message.
        subcommand = self.make_subcommand(help='help message')
        parser = cli.get_parser([('foo', subcommand)])
        help = parser.format_help()
        self.assertIn(subcommand.help, help)
        self.assertIn('foo', help)

    def test_help_subcommand(self):
        # Ensure the help subcommand correctly show help messages retrieving
        # info from other subcommands.
        name = 'foo'
        subcommand = self.make_subcommand(
            help='subcommand help message',
            add_arguments=lambda self, parser: parser.add_argument('--foo'))
        parser = cli.get_parser([(name, subcommand)])
        namespace = parser.parse_args(['help', name])
        with self.assertRaises(SystemExit) as cm:
            with capture_output() as output:
                namespace.main(namespace)
        output_value = output.getvalue()
        # The help subcommand exits without errors.
        self.assertEqual(0, cm.exception.code)
        # The subcommand help is present in the output.
        self.assertIn(subcommand.help, output_value)
        # The subcommand name is displayed as well.
        self.assertIn(name, output_value)
        # The help subcommand also prints arguments info.
        self.assertIn('--foo', output_value)

    def test_epilog_in_global_help(self):
        # The epilog attribute of subcommands is used to generate
        # the base command usage message.
        subcommand = self.make_subcommand(epilog='epilog message')
        parser = cli.get_parser([('foo', subcommand)])
        help = parser.format_help()
        self.assertIn(subcommand.epilog, help)

    def test_subcommand_epilog(self):
        # The epilog attribute of subcommands is used to create the parser,
        # so that the epilog text is included in the subcommand help.
        name = 'foo'
        subcommand = self.make_subcommand(epilog='epilog message')
        parser = cli.get_parser([(name, subcommand)])
        namespace = parser.parse_args(['help', name])
        with self.assertRaises(SystemExit):
            with capture_output() as output:
                namespace.main(namespace)
        self.assertIn(subcommand.epilog, output.getvalue())

    def test_subcommands(self):
        # Ensure sub parsers are added for each registered subcommand.
        subcommands = (
            ('foo', self.make_subcommand_returning('foo')),
            ('bar', self.make_subcommand_returning('bar')),
            )
        parser = cli.get_parser(subcommands)
        for name in dict(subcommands):
            namespace = parser.parse_args([name])
            self.assertEqual(name, namespace.main(namespace))


class PrepareParserTestMixin(ParserTestMixin, SubCommandFactoryTestMixin):

    def prepare_and_parse(self, subcommand, args=()):
        """Prepare and parse *self.parser* passing given *subcommand*.

        Return the resulting namespace.
        """
        cli.prepare_parser(self.parser, subcommand)
        return self.parser.parse_args(args)


class PrepareParserTest(PrepareParserTestMixin, unittest.TestCase):

    def test_main_function_stored(self):
        # Ensure *prepare_parser* stores the main function in the parser
        # defaults, and that the main function calls *subcommand.run*.
        subcommand = self.make_subcommand_returning('foo')
        namespace = self.prepare_and_parse(subcommand)
        self.assertEqual('foo', namespace.main(namespace))

    def test_common_arguments(self):
        # Ensure common arguments are added to the parser and to the resulting
        # namespace.
        subcommand = self.make_subcommand(
            has_dry_run=True, has_interactive_run=True)
        args = ('--dry-run', '--yes')
        namespace = self.prepare_and_parse(subcommand, args)
        self.assertTrue(namespace.dry_run)
        self.assertFalse(namespace.interactive)

    def test_subcommand_arguments(self):
        # Ensure subcommand specific arguments are added to the parser as well.
        subcommand = self.make_subcommand(
            add_arguments=lambda self, parser: parser.add_argument('--foo'))
        args = ('--foo', 'bar')
        namespace = self.prepare_and_parse(subcommand, args)
        self.assertEqual('bar', namespace.foo)


class RunTest(PrepareParserTestMixin, unittest.TestCase):

    def assertHasAttr(self, obj, attr):
        """Assert given *obj* has *attr*."""
        self.assertTrue(
            hasattr(obj, attr),
            '%r does not have %r attribute' % (obj, attr))

    def make_described_subcommand(self, **kwargs):
        """Create a subcommand having a *get_description* method.

        The created subcommand returns None.
        """
        get_description = lambda self, namespace: 'subcommand description'
        return self.make_subcommand_returning(
            None, get_description=get_description, **kwargs)

    def test_init_namespace(self):
        # Ensure *argparser.init_namespace* is invoked by *run*.
        subcommand = self.make_subcommand_returning('foo')
        namespace = self.prepare_and_parse(subcommand)
        cli.run(self.parser, subcommand, namespace)
        self.assertHasAttr(namespace, 'euid')
        self.assertHasAttr(namespace, 'run_as_root')

    def test_prepare_namespace(self):
        # Ensure *subcommand.prepare_namespace*, if defined, is invoked.
        def prepare_namespace(self, namespace):
            namespace.prepared = True

        subcommand = self.make_subcommand_returning(
            'foo', prepare_namespace=prepare_namespace)
        namespace = self.prepare_and_parse(subcommand)
        cli.run(self.parser, subcommand, namespace)
        self.assertTrue(namespace.prepared)

    def test_validation_error_in_prepare_namespace(self):
        # Ensure the execution is stopped if *prepare_namespace* raises a
        # ValidationError.
        msg = 'Nothing is really going on.'

        def prepare_namespace(self, namespace):
            raise exceptions.ValidationError(msg)

        subcommand = self.make_subcommand(prepare_namespace=prepare_namespace)
        namespace = self.prepare_and_parse(subcommand)
        with self.assertRaises(SystemExit) as cm:
            with capture_error() as error:
                cli.run(self.parser, subcommand, namespace)
        # The program exits with an error.
        self.assertEqual(2, cm.exception.code)
        # The validation error is printed to stderr.
        self.assertIn(msg, error.getvalue())

    def test_dry_run(self):
        # The subcommand's description is printed to stdout and the execution
        # is stopped without errors if this is a dry run.
        subcommand = self.make_described_subcommand(has_dry_run=True)
        namespace = self.prepare_and_parse(subcommand, ['--dry-run'])
        with capture_output() as output:
            retcode = cli.run(self.parser, subcommand, namespace)
        # The command description is displayed.
        self.assertIn(subcommand.get_description(namespace), output.getvalue())
        # The command exited without errors.
        self.assertIsNone(retcode)

    def test_dry_run_is_never_interactive(self):
        # Ensure a dry run is never interactive.
        subcommand = self.make_described_subcommand(
            has_dry_run=True, has_interactive_run=True)
        namespace = self.prepare_and_parse(subcommand, ['--dry-run'])
        with nested(capture_output(), RawInputReturning('')) as cms:
            cli.run(self.parser, subcommand, namespace)
        # The confirm function has not been called.
        self.assertEqual(0, cms[1].call_count)

    def test_dry_run_never_requires_root(self):
        # Ensure a dry run never requires to restart as root.
        subcommand = self.make_described_subcommand(
            root_required=True, has_dry_run=True)
        namespace = self.prepare_and_parse(subcommand, ['--dry-run'])
        with capture_output():
            retcode = cli.run(self.parser, subcommand, namespace)
        # The command exited without errors and without a restart as root.
        self.assertIsNone(retcode)

    def test_interactive_execution_granted(self):
        # Ensure the execution continues if the user confirms to proceed.
        subcommand = self.make_subcommand_returning(
            None, has_interactive_run=True)
        namespace = self.prepare_and_parse(subcommand)
        with nested(capture_output(), RawInputReturning('yes')) as cms:
            retcode = cli.run(self.parser, subcommand, namespace)
        # The command exited without errors.
        self.assertIsNone(retcode)
        # The user confirmed the execution.
        self.assertEqual(1, cms[1].call_count)

    def test_interactive_execution_denied(self):
        # Ensure the command exits with an error if the user denies execution.
        subcommand = self.make_subcommand_returning(
            None, has_interactive_run=True)
        namespace = self.prepare_and_parse(subcommand)
        with nested(capture_output(), RawInputReturning('no')):
            retcode = cli.run(self.parser, subcommand, namespace)
        self.assertEqual(1, retcode)

    def test_assume_yes(self):
        # Ensure confirmation is not asked if `--yes` is provided.
        subcommand = self.make_subcommand_returning(
            None, has_interactive_run=True)
        namespace = self.prepare_and_parse(subcommand, ['--yes'])
        with nested(capture_output(), RawInputReturning('')) as cms:
            retcode = cli.run(self.parser, subcommand, namespace)
        # The confirm function has not been called.
        self.assertEqual(0, cms[1].call_count)
        # The command exited without errors.
        self.assertIsNone(retcode)

    def test_restart_as_root(self):
        # Ensure the execution is restarted as root if required.
        subcommand = self.make_described_subcommand(root_required=True)
        namespace = self.prepare_and_parse(subcommand)
        with call_replaced_by(lambda cmd: cmd):
            cmd = cli.run(self.parser, subcommand, namespace)
        self.assertEqual('sudo', cmd[0])

    def test_execution_error_in_run(self):
        # An error is returned if the subcommand's *run()*
        # raises an ExecutionError
        execution_error = exceptions.ExecutionError()

        def run(self, namespace):
            raise execution_error

        subcommand = self.make_subcommand(run=run)
        namespace = self.prepare_and_parse(subcommand)
        error = cli.run(self.parser, subcommand, namespace)
        self.assertIs(execution_error, error)


class MainTest(PrepareParserTestMixin, unittest.TestCase):

    def test_normal_execution(self):
        # Ensure *main* executes the main function as it is found
        # in the namespace, and return its return value.
        self.parser.set_defaults(main=lambda namespace: 1)
        self.assertEqual(1, cli.main([], self.parser))

    def test_subcommand_execution(self):
        # Ensure *main* returns the subcommand's *run()* return value.
        subcommand = self.make_subcommand_returning(1)
        parser = cli.get_parser([('foo', subcommand)])
        self.assertEqual(1, cli.main(['foo'], parser))

    def test_help(self):
        # Ensure *main* returns a non-error exit code if help is invoked.
        with capture_output():
            self.assertEqual(0, cli.main(['-h'], self.parser))

    def test_subcommand_help(self):
        # Ensure *main* returns a non-error exit code if a subcommand's help
        # is requested.
        subcommand = self.make_subcommand(help='help message')
        parser = cli.get_parser([('foo', subcommand)])
        with capture_output() as output:
            retcode = cli.main(['help', 'foo'], parser)
        self.assertEqual(0, retcode)
        self.assertIn(subcommand.help, output.getvalue())

    def test_invalid_arguments(self):
        # Ensure *main* returns an error exit code if wrong arguments are
        # provided.
        invalid_args = 'invalid arguments'
        with capture_error() as error:
            retcode = cli.main(invalid_args.split(), self.parser)
        self.assertEqual(2, retcode)
        self.assertIn(invalid_args, error.getvalue())
