#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Test utilities."""

__metaclass__ = type
__all__ = [
    'call_replaced_by',
    'capture',
    'capture_error',
    'capture_output',
    'create_test_branch',
    'get_random_string',
    'ParserTestMixin',
    'RawInputReturning',
    'skip_if_no_lpuser',
    'StepsFactoryTestMixin',
    'SubCommandFactoryTestMixin',
    ]

from contextlib import contextmanager
from functools import partial
import os
import random
import shutil
import string
from StringIO import StringIO
import sys
import subprocess
import tempfile
import unittest

from shelltoolbox import (
    environ,
    run,
    )

from lpsetup import argparser
from lpsetup.utils import call


@contextmanager
def call_replaced_by(function):
    """Temporarily mock *subprocess.call* with the given *function*."""
    original, subprocess.call = subprocess.call, function
    try:
        yield
    finally:
        subprocess.call = original


class CallReplacedByTest(unittest.TestCase):
    """Tests for the *call_replaced_by* context manager."""

    def test_call_is_replaced(self):
        # Ensure *subprocess.call* is correctly replaced.
        with call_replaced_by(lambda *args, **kwargs: 'mocked'):
            self.assertEqual('mocked', subprocess.call(['ls']))

    def test_call_is_restored(self):
        # Ensure *subprocess.call* is correctly restored exiting from the cm.
        original = subprocess.call
        with call_replaced_by(lambda *args, **kwargs: 'mocked'):
            pass
        self.assertIs(original, subprocess.call)


@contextmanager
def capture(attr):
    output = StringIO()
    backup = getattr(sys, attr)
    setattr(sys, attr, output)
    try:
        yield output
    finally:
        setattr(sys, attr, backup)


capture_output = partial(capture, 'stdout')
capture_error = partial(capture, 'stderr')


class CaptureTest(unittest.TestCase):
    """Tests for the *capture* context manager."""

    message = 'message'

    def test_capture_output(self):
        with capture_output() as stream:
            print self.message
        self.assertEqual(self.message + '\n', stream.getvalue())

    def test_capture_error(self):
        with capture_error() as stream:
            print >> sys.stderr, self.message
        self.assertEqual(self.message + '\n', stream.getvalue())


# Template directory used to create a Launchpad branch.
test_template_dir = os.path.join(os.path.dirname(__file__), 'test-branch')


def create_test_branch(template_dir=None):
    """Create a temporary test branch containing files in *template_dir*.

    If *template_dir* is None, the files in `lpsetup/tests/test-branch`
    are used.

    Return the path of the newly created branch.
    """
    if template_dir is None:
        template_dir = test_template_dir
    branch_path = tempfile.mktemp()
    shutil.copytree(template_dir, branch_path)
    call('bzr', 'init', '--quiet', branch_path)
    call('bzr', 'add', '--quiet', branch_path)
    call('bzr', 'commit', branch_path, '-m', 'Initial commit.')
    return branch_path


class CreateTestBranchTest(unittest.TestCase):
    """Tests for the *create_test_branch* helper function."""

    def setUp(self):
        # Create a Bazaar user id inside a temporary home.
        self.temp_home = tempfile.mkdtemp()
        with environ(HOME=self.temp_home):
            run('bzr', 'whoami', 'Example User <user@example.com>')
            self.branch_path = create_test_branch()

    def tearDown(self):
        shutil.rmtree(self.temp_home)
        shutil.rmtree(self.branch_path)

    def test_revno(self):
        # Ensure the bzr branch is correctly created and contains a revision.
        revno = run('bzr', 'revno', self.branch_path)
        self.assertEqual(1, int(revno.strip()))

    def test_working_tree(self):
        # Ensure the working tree is present in the newly created branch.
        test_file_path = os.path.join(self.branch_path, 'test-file')
        self.assertTrue(os.path.exists(test_file_path))


def get_random_string(size=10):
    """Return a random string to be used in tests."""
    return ''.join(random.sample(string.ascii_letters, size))


# Retrieve the current lpuser to be used in tests.
try:
    lpuser = run('bzr', 'launchpad-login').strip()
except subprocess.CalledProcessError:
    lpuser = None
# Create a decorator to skip tests if lp login is not set.
skip_if_no_lpuser = unittest.skipIf(
    lpuser is None, 'You need to set up a Launchpad login to run this test.')


class ParserTestMixin(object):

    def setUp(self):
        """Set up an argument parser."""
        self.parser = argparser.ArgumentParser()


class RawInputReturning(object):
    """Mocks the *raw_input* built-in function.

    This context manager takes one or more pre-defined answers and
    keeps track of mocked *raw_input* call count.
    """
    def __init__(self, *args):
        self.answers = iter(args)
        self._builtin = __import__('__builtin__')
        self.call_count = 0
        self.original = self._builtin.raw_input

    def __enter__(self):
        self._builtin.raw_input = self.input
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._builtin.raw_input = self.original

    def input(self, prompt=''):
        self.call_count += 1
        return self.answers.next()


class RawInputReturningTest(unittest.TestCase):
    """Tests for the *RawInputReturning* context manager."""

    def test_original_is_replaced(self):
        # Ensure the original function is correctly replaced.
        with RawInputReturning('mocked'):
            self.assertEqual('mocked', raw_input())

    def test_original_is_retrievable(self):
        # Ensure the original function is still stored as an attribute of the
        # context manager.
        original = raw_input
        with RawInputReturning('mocked') as cm:
            self.assertIs(original, cm.original)

    def test_multiple_return_values(self):
        # Each call returns one of the provided values in order.
        with RawInputReturning(*range(2)):
            self.assertEqual(0, raw_input())
            self.assertEqual(1, raw_input())

    def test_call_count(self):
        # Ensure the context manager keeps track of call count.
        return_values = range(2)
        with RawInputReturning(*return_values) as cm:
            self.assertEqual(0, cm.call_count)
            for i in return_values:
                raw_input()
            self.assertEqual(2, cm.call_count)


class StepsFactoryTestMixin(object):

    def make_step(self, name='step'):
        """Return a step callable."""
        def step(*args):
            return args

        step.__name__ = name
        return step

    def make_named_step(self, step_name):
        """Return a named step callable."""
        step = self.make_step()
        step.step_name = step_name
        return step

    def make_described_step(self, description):
        """Return a step having a description."""
        step = self.make_step()
        step.description = description
        return step


class SubCommandFactoryTestMixin(object):

    def make_subcommand(self, **kwargs):
        """Create and return a subcommand.

        Attributes can be specified using kwargs.
        """
        return type('SubCommand', (argparser.BaseSubCommand,), kwargs)()

    def make_subcommand_returning(self, value, **kwargs):
        """Create a subcommand having the run function returning *value*."""
        run = lambda self, namespace: value
        return self.make_subcommand(run=run, **kwargs)

    def make_steps_based_subcommand(self, **kwargs):
        """Create and return a steps based subcommand.

        Attributes can be specified using kwargs.
        """
        return type(
            'StepsBasedSubCommand',
            (argparser.StepsBasedSubCommand,),
            kwargs)()
